
"""SQL script locations."""

__all__ = ['sql_locations']

sql_locations = [
    "/usr/share/skytools3",
]

# PEP 440 version: [N!]N(.N)*[{a|b|rc}N][.postN][.devN]

package_version = "3.9.2"

skylog = 0

